/*  plumb - the pipe shell - diagnostics, debug, dump
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <stdarg.h>
#include "genht/htsp.h"
#include "genht/hash.h"
#include "process.h"
#include "vp.h"
#include "rp.h"
#include "plumb.h"
#include "config.h"

/* these shouldn't end up in .h files as they are really internals for
   the corresponding module; still, a central diagnostics needs to access
   then */
extern htsp_t *procs;
extern fd_t **wfds;
extern struct P_pollfd *pfds;
extern int wfds_alloced, wfds_used;
extern pipe_t *pipes_head;


struct pbit_s {
	int bitval;
	char *name;
};

static struct pbit_s pbits[] = {
	{P_POLLIN,  "POLLIN"},
	{P_POLLOUT, "POLLOUT"},
	{0, NULL}
};

#define pbool(b)   ((b) ? "yes" : "no")

const char *query_proc_state(proc_virt_t *pv)
{
	switch(pv->h.state) {
		case PS_STOPPED: return "stopped"; break;
		case PS_RUNNING: return "running"; break;
		case PS_PAUSED:  return "paused"; break;
	}
	return "<unknown>";
}

static const char *proc_id(fd_t *fd)
{
	if (fd == NULL)
		return "-";
	return fd->proc.v->h.id;
}

static int proc_fd(fd_t *fd)
{
	if (fd == NULL)
		return -1;
	return fd->fd;
}

void print_internals(FILE *f, const char *label)
{
	htsp_entry_t *e;
	proc_virt_t *pv;
	proc_real_t *pr;
	int n;
	pipe_t *p;


	if (f == NULL)
		f = stderr;
	fprintf(f, "### start of plumb internals\n");
	fprintf(f, "# label %s\n", label);
	fprintf(f, "# watches:\n");
	for(n = 0; n < wfds_used; n++) {
		struct pbit_s *b;
		int first;

		fprintf(f, "watch/%d %s/%d ", n, wfds[n]->proc.v->h.id, wfds[n]->fd);
		for(first = 1, b = pbits; b->name != NULL; b++) {
			if (pfds[n].events & b->bitval) {
				if (first)
					first = 0;
				else
					fprintf(f, "|");
				fprintf(f, "%s", b->name);
			}
		}
		fprintf(f, "\n");
	}

	fprintf(f, "# processes:\n");
	for (e = htsp_first(procs); e; e = htsp_next(procs, e)) {
		pv = e->value;
		pr = e->value;
		fprintf(f, "proc %s (%s)\n", pv->h.id, (pv->h.virt ? "virt" : "real"));
		fprintf(f, "proc/transaction %s\n", pv->h.tx_name == NULL ? "<unknown>" : pv->h.tx_name);
		fprintf(f, "proc/state %s\n", query_proc_state(pv));
		fprintf(f, "proc/fds\n");
		for(n = 0; n < pv->h.num_fds; n++) {
			char *role;
			if (is_sink(&pv->h.fds[n]))
				role = "sink";
			else if (is_source(&pv->h.fds[n]))
				role = "source";
			else
				role = "NONE";
			fprintf(f, "proc/fd/%d %s %p %s %s %s %s", n, role, pv->h.fds[n].pipe, pv->h.fds[n].eof ? "eof" : "-", pv->h.fds[n].blocked ? "blocked" : "-", pv->h.fds[n].closing ? "closing" : "-", pv->h.fds[n].paused ? "paused" : "-" );
			if (!pv->h.virt)
				fprintf(f, " %d", pr->sock[n]);
			fprintf(f, "\n");
		}
		if (!pv->h.virt) {
			char **s;
			fprintf(f, "proc/args\n");
			for(n = 0, s = pr->argv; *s != NULL; s++,n++)
				fprintf(f, "proc/arg/%d %s\n", n, *s);
			if (pr->shadow_of != NULL)
				fprintf(f, "proc/shadow_of %s\n", pr->shadow_of->h.id);
		}
		else {
			fprintf(f, "proc/args\n");
			fprintf(f, "proc/arg/0 %s\n", pv->cmd);
		}
		fprintf(f, "proc/properties\n");
		fprintf(f, "proc/property/sticky %s\n", pbool(pr->h.sticky));
		fprintf(f, "proc/property/respwan %s\n", pbool(pr->h.respawn));
		fprintf(f, "proc/property/kill_on_eof %s\n", pbool(pr->h.kill_on_eof));
	}

	fprintf(f, "# pipes:\n");
	for(p = pipes_head; p != NULL; p = p->next)
		fprintf(f, "pipe/%p %s:%d -> %s:%d\n", p, proc_id(p->from), proc_fd(p->from), proc_id(p->to), proc_fd(p->to));

	fprintf(f, "### end of plumb internals\n");
	fflush(f);
}

static int in_error = 0;
void plumb_error(char *fmt, ...)
{
	va_list ap;
	FILE *f;

	plumb_shutdown = 1;
	fprintf(stderr, "plumb: error: ");
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	if (in_error)
		return;
	in_error=1;
	f = fopen("plumb.crash", "w");
	if (f != NULL) {
		print_internals(f, "plumb crash");
		fclose(f);
		fprintf(stderr, "Map of internals is saved in plumb.crash\n");
	}

	plumb_stop_RPs();
	abort();
}

static char *eventb = NULL;
int eventb_alloced = 0;
fd_t *eventfd = NULL;

void plumb_event(int maxlen, char *fmt, ...)
{
	va_list ap;
	int len;

	/* don't deliver events during shutdown */
	if (plumb_shutdown)
		return;

	/* if needed reallocate buffer to match expected max length */
	if (maxlen >= eventb_alloced) {
		eventb_alloced = maxlen + 128;
		eventb = realloc(eventb, eventb_alloced);
	}

	/* bind an event fd if needed */
	if (eventfd == NULL)
		eventfd = fd_ref("plumb_core", 1);

	va_start(ap, fmt);
	len = vsprintf(eventb, fmt, ap);
	va_end(ap);

	fd_write(eventfd, eventb, len);
}

void plumb_event_uninit(void)
{
	if (eventfd != NULL) {
		fd_close(eventfd);
		eventfd = NULL;
	}
	if (eventb != NULL) {
		free(eventb);
		eventb_alloced = 0;
		eventb = NULL;
	}
}

static void query_out(fd_t *fd, const char *msg)
{
	fd_write(fd, msg, strlen(msg));
}

static void query_out_start(fd_t *fd, char *cookie, int success, const char *msg)
{
	if (success)
		fd_write(fd, "query-ans [", 11);
	else
		fd_write(fd, "query-err [", 11);

	if (cookie != NULL)
		fd_write(fd, cookie, strlen(cookie));

	if (msg != NULL) {
		fd_write(fd, "] ", 2);
		query_out(fd, msg);
	}
	else
		fd_write(fd, "]", 1);
}

static void query_out_end(fd_t *fd)
{
	fd_write(fd, "\n", 1);
}

void run_query(fd_t *out, char *q, char *object, char *cookie)
{
	if (strncmp(q, "proc.", 5) == 0) {
		proc_virt_t *pv, *p;
		proc_real_t *pr;
		q += 5;
		if (proc_lookup(object, &pv, &pr) < 0) {
			query_out_start(out, cookie, 0, "process does not exist: ");
			query_out(out, object);
			query_out_end(out);
			return;
		}
		p = (pv == NULL) ? (proc_virt_t *)pr : pv; /* we need p for the headers */
		if (strcmp(q, "state") == 0) {
			query_out_start(out, cookie, 1, query_proc_state(p));
			query_out_end(out);
			return;
		}
	}
	query_out_start(out, cookie, 0, "error invalid query: unknown property ");
	query_out(out, q);
	query_out_end(out);
}


#ifdef RW_DEBUG

void fprintchr(FILE *f, char c)
{
	switch(c) {
		case '\n': fputc('\\', f);fputc('n', f); break;
		case '\r': fputc('\\', f);fputc('r', f); break;
		case '\t': fputc('\\', f);fputc('t', f); break;
		default:
			if (c < 32)
				fputc('~', f);
			else
				fputc(c, f);
	}
}

void fprintstr(FILE *f, char *str)
{
	char *s, *start;
	int n, size = 8;

	/* print first size chars */
	fputc('\"', f);
	for(s = str, n = 0; (*s != '\0') && (n < size); s++, n++)
		fprintchr(f, *s);

	/* search for the last size chars */
	for(start = s; (*s != '\0'); s++) ;
	s -= size;

	/* the string is long enough to be split */
	if (s > start)
		fprintf(f, "\" ... \"");

	/* print the second portion */
	for(; (*s != '\0'); s++)
		fprintchr(f, *s);
	fputc('\"', f);
}
#endif
