/*  plumb - the pipe shell - helpers for real processes (RPs)
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libporty/net_async/abgproc.h>
#include "process.h"
#include "pipe.h"
#include "plumb.h"
#include "rp.h"

/* defined in virt/procwatch.c; shorthand for easier binding of control socket */
extern proc_virt_t *procwatch;

static int rp_init(proc_real_t *p)
{
	fprintf(stderr, "TODO: start process %p\n", p);
	return -1;
}

proc_real_t *rp_alloc(transaction_t *tx, char *cmd, const char *id, kvlist_t *args)
{
	proc_real_t *p;
	kv_t *kv;
	int n, len;

	/* set up header with default properties */
	p = calloc(sizeof(proc_virt_t), 1);
	p->h.virt        = 0;
	p->h.id          = strclone(id);
	if (tx != NULL)
		p->h.tx_name   = strclone(tx->name);
	p->h.sticky      = 1;
	p->h.respawn     = 0;
	p->h.kill_on_eof = 0;
	p->init          = rp_init;

	/* make sure comm fds are -1 to avoiding writing to fd 0 */
	p->sock_ctrl   = -1;

	/* construct argv */
	if (args != NULL)
		len = (args->len + 2);
	else
		len = 2;

	p->argv = malloc(sizeof(char *) * (len+2));
	p->argv[0] = cmd;
	if (args != NULL) {
		for(n = 0, kv = args->head; n < args->len; n++, kv = kv->next) {
			p->argv[n+1] = kv->key;
			kv->key = NULL; /* kv will be free'd later, we don't want to lose this string */
		}
		p->argv[n+1] = NULL;
	}
	else
		p->argv[1] = NULL;

	return p;
}

void rp_free(proc_real_t *p)
{
	char **a;
	if (p->sock != NULL)
		free(p->sock);
	for(a = p->argv; *a != NULL; a++)
		free(*a);
	free(p->argv);
}


void rp_start(proc_real_t *pr)
{
	int ret, n, last;
	P_net_socket *child_sockets;
	int *child_fds;
	fd_t *ctrlfd, *watchfd;

#warning TODO: inherit more fds when needed
	child_sockets = malloc(sizeof(P_net_socket) * pr->h.num_fds);
	child_fds     = malloc(sizeof(int) * pr->h.num_fds);
	last = 0;
	for(n = 0; n < pr->h.num_fds; n++) {
		fd_t *fd = &pr->h.fds[n];
		if ((is_source(fd)) || (is_sink(fd))) {
			if (P_net_local_socket(&pr->sock[n], &child_sockets[last], P_net_nonblock_full) != 0)
				plumb_error("can't create pipe for fd %d of process %s (%s)\n", n, pr->h.id, pr->argv[0]);
			child_fds[last] = n;
			last++;
		}
	}

	ret = P_net_async_bgproc_inherit(pr->argv, &pr->sock_ctrl, child_sockets, child_fds, last);

	/* close the child-side of sockets in the parent */
	for(n = 0; n < last; n++)
		close(child_sockets[n]);

	free(child_sockets);
	free(child_fds);

	if (ret != 0)
		plumb_error("can't start process %s (%s)\n", pr->h.id, pr->argv[0]);

	for(n = 0; n < pr->h.num_fds; n++) {
		fd_t *fd = &pr->h.fds[n];
		if (is_source(fd)) {
			fd_set_socket(fd, pr->sock[n]);
#warning TODO: call fd_rp_src_poll()
			fd_add_watch(fd, P_POLLIN);
		}
		else if (is_sink(fd)) {
			fd_set_socket(fd, pr->sock[n]);
			/* no need to add it as P_POLLOUT until the first write fails */
		}
	}

	/* bind control socket to procwatch; ctrlfd is pr's next free fd (but >=3)
	   as source, watchfd s procswatch's next free socket as sink */
	ctrlfd  = fd_ref_((proc_virt_t*)pr, (pr->h.num_fds < 3) ? 3 : -1);
	watchfd = fd_ref_(procwatch, -1);
	fd_set_socket(ctrlfd, pr->sock_ctrl);
	fd_add_watch(ctrlfd, P_POLLIN);
	pipe_register(ctrlfd, watchfd);

	/* mark it running */
	pr->h.state = PS_RUNNING;
}

void rp_stop(proc_real_t *pr, int hard)
{
	/* real process: just kill it using libporty */

	if (pr->stop != NULL)
		pr->stop(pr, hard);

	if (pr->sock_ctrl > 0) {
		if (hard)
			P_net_write(pr->sock_ctrl, "KILL\n", 5);
		else
			P_net_write(pr->sock_ctrl, "TERM\n", 5);
		/* no need to change anything in the structs - _procwatch will be notified
		   when the process really dies and cleans up */
	}
}

void rp_socket_inval(proc_real_t *pr, P_net_socket socket)
{
	int n;

	if (socket == pr->sock_ctrl)
		pr->sock_ctrl = P_NET_INVALID_SOCKET;
	else {
		for(n = 0; n < pr->h.num_fds; n++) {
			if (socket == pr->sock[n])
				pr->sock[n] = P_NET_INVALID_SOCKET;
		}
	}
}
